/*
 * Portions of this software are copyrighted by the 
 * 
 *     University of South Carolina,
 *     Sun Microsystems, Inc., 
 * 
 * and other parties. 
 * 
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
 * DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
 * IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
 * NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 */


package edu.sc.seis.TauP;

/** convenience class for storing the parameters associated with a
  * phase arrival. 
  *
  * @version 1.0  Tue Feb  2 20:48:34 GMT 1999
  * @author H. Philip Crotwell
  *
  */
public class Arrival implements Cloneable {

		/** travel time in seconds */
   protected double time;

		/** angular distance (great circle) in radians */
   protected double dist;

		/** ray parameter in seconds per radians. */
   protected double rayParam;

   protected int rayParamIndex;

		/** phase name */
   protected String name;
   
   	/** phase name changed for true depths */
   protected String puristName;

		/** source depth in kilometers */
   protected double sourceDepth;

		/** pierce and path points */
   protected TimeDist[] pierce, path;

// get set methods

		/** returns travel time in seconds */
	public double getTime() {
		return time;
	}

		/** returns travel distance in radians */
	public double getDist() {
		return dist;
	}

		/** returns ray parameter in seconds per radian */
	public double getRayParam() {
		return rayParam;
	}

		/** returns phase name */
	public String getName() {
		return name;
	}

		/** returns purist's version of name. Depths are changed to reflect
		  * the true depth of the interface. */
	public String getPuristName() {
		return puristName;
	}

		/** returns source depth in kilometers */
	public double getSourceDepth() {
		return sourceDepth;
	}

		/** returns pierce points as TimeDist objects. */
	public TimeDist[] getPierce() {
		return pierce;
	}

		/** returns pierce points as TimeDist objects. */
	public TimeDist[] getPath() {
		return path;
	}

   public Object clone() {
      try {
         Arrival newObject = (Arrival)super.clone();
         if (pierce != null) {
            newObject.pierce = new TimeDist[pierce.length];
            for (int i=0; i<pierce.length; i++) {
               if (pierce[i] != null) {
                  newObject.pierce[i] = (TimeDist)pierce[i].clone();
               } else {
                  newObject.pierce[i] = null;
               }
            }
         }
         if (path != null) {
            newObject.path = new TimeDist[path.length];
            for (int i=0; i<path.length; i++) {
               newObject.path[i] = (TimeDist)path[i].clone();
            }
         }
         return newObject;
      } catch (CloneNotSupportedException e) {
         // Can't happen, but...
         throw new InternalError(e.toString());
      }
   }

   public String toString() {
      String desc = "time="+time+" dist="+dist+" rayParam="+rayParam+" rayParamIndex="+rayParamIndex+" name="+name;
		if (pierce != null) {
			System.out.println("\nPierce:");
			for (int i=0;i<pierce.length;i++) {
				System.out.println(pierce[i]);
			}
		}
      return desc;
   }

	public int getNumPiercePoints() {
		if (pierce != null) {
			return pierce.length;
		} else {
			return 0;
		}
	}

	public int getNumPathPoints() {
		if (path != null) {
			return path.length;
		} else {
			return 0;
		}
	}

	public TimeDist getPiercePoint(int i) {
			// don't check for i> length since we want an ArrayOutOfBounds anyway
		return pierce[i];
	}

		/** finds the first pierce point at the given depth.
		 *  @throws ArrayIndexOutOfBoundsException if depth is not found
		 */
	public TimeDist getFirstPiercePoint(double depth) {
		for (int i=0; i< pierce.length; i++) {
			if (pierce[i].depth == depth) {
				return pierce[i];
			}
		}
		throw new ArrayIndexOutOfBoundsException(
			"No Pierce point found for depth "+depth);
	}

		/** finds the last pierce point at the given depth.
		 *  @throws ArrayIndexOutOfBoundsException if depth is not found
		 */
	public TimeDist getLastPiercePoint(double depth) {
		TimeDist piercepoint = null;
		for (int i=0; i< pierce.length; i++) {
			if (pierce[i].depth == depth) {
				piercepoint =  pierce[i];
			}
		}
		if (piercepoint == null) {
			throw new ArrayIndexOutOfBoundsException(
				"No Pierce point found for depth "+depth);
		}
		return piercepoint;
	}

	public TimeDist getPathPoint(int i) {
			// don't check for i> length since we want an ArrayOutOfBounds anyway
		return path[i];
	}

}
